@@ -297,7 +297,7 @@ archive_file_get_from_path (ArchiveFile *file, const char *filename, gboolean ad
 static void
 create_root_file (GVfsBackendArchive *ba)
 {
-  ArchiveFile *root = g_slice_new0 (ArchiveFile);
+  ArchiveFile *root;
   GFileInfo *info;
   char *s, *display_name;
   GIcon *icon;
@@ -583,7 +583,9 @@ do_mount (GVfsBackend *backend,
 
 static void
 do_unmount (GVfsBackend *backend,
-	    GVfsJobUnmount *job)
+	    GVfsJobUnmount *job,
+            GMountUnmountFlags flags,
+            GMountSource *mount_source)
 {
   GVfsBackendArchive *ba = G_VFS_BACKEND_ARCHIVE (backend);
 
