@@ -95,6 +95,13 @@ static char *default_workgroup = NULL;
 
 G_DEFINE_TYPE (GVfsBackendSmb, g_vfs_backend_smb, G_VFS_TYPE_BACKEND)
 
+static void set_info_from_stat (GVfsBackendSmb *backend,
+				GFileInfo *info,
+				struct stat *statbuf,
+				const char *basename,
+				GFileAttributeMatcher *matcher);
+
+
 static void
 g_vfs_backend_smb_finalize (GObject *object)
 {
@@ -653,6 +660,15 @@ try_mount (GVfsBackend *backend,
   return FALSE;
 }
 
+static int
+fixup_open_errno (int err)
+{
+  /* samba has a bug (#6228) where it doesn't set errno if path resolving failed */
+  if (err == 0)
+    err = ENOTDIR;
+  return err;
+}
+
 static void 
 do_open_for_read (GVfsBackend *backend,
 		  GVfsJobOpenForRead *job,
@@ -670,11 +686,13 @@ do_open_for_read (GVfsBackend *backend,
 
   uri = create_smb_uri (op_backend->server, op_backend->share, filename);
   smbc_open = smbc_getFunctionOpen (op_backend->smb_context);
+  errno = 0;
   file = smbc_open (op_backend->smb_context, uri, O_RDONLY, 0);
 
   if (file == NULL)
     {
-      olderr = errno;
+      olderr = fixup_open_errno (errno);
+      
       smbc_stat = smbc_getFunctionStat (op_backend->smb_context);
       res = smbc_stat (op_backend->smb_context, uri, &st);
       g_free (uri);
@@ -705,12 +723,14 @@ do_read (GVfsBackend *backend,
   ssize_t res;
   smbc_read_fn smbc_read;
 
-  /* For some reason requests of 65536 bytes broke for me (returned 0)
-   * Maybe some smb protocol limit
+  /* libsmbclient limits blocksize to (64*1024)-2 for Windows servers,
+   * let's do the same here to achieve reasonable performance. (#588391)
+   *
+   * TODO: port to pull mechanism (#592468)
    */
-  if (bytes_requested > 65535)
-    bytes_requested = 65535;
-  
+  if (bytes_requested > 65534)
+    bytes_requested = 65534;
+
   smbc_read = smbc_getFunctionRead (op_backend->smb_context);
   res = smbc_read (op_backend->smb_context, (SMBCFILE *)handle, buffer, bytes_requested);
 
@@ -720,7 +740,6 @@ do_read (GVfsBackend *backend,
     {
       g_vfs_job_read_set_size (job, res);
       g_vfs_job_succeeded (G_VFS_JOB (job));
-	    
     }
 }
 
@@ -768,6 +787,33 @@ do_seek_on_read (GVfsBackend *backend,
   return;
 }
 
+static void
+do_query_info_on_read (GVfsBackend *backend,
+		       GVfsJobQueryInfoRead *job,
+		       GVfsBackendHandle handle,
+		       GFileInfo *info,
+		       GFileAttributeMatcher *matcher)
+{
+  GVfsBackendSmb *op_backend = G_VFS_BACKEND_SMB (backend);
+  struct stat st = {0};
+  int res, saved_errno;
+  smbc_fstat_fn smbc_fstat;
+
+  smbc_fstat = smbc_getFunctionFstat (op_backend->smb_context);
+  res = smbc_fstat (op_backend->smb_context, (SMBCFILE *)handle, &st);
+  saved_errno = errno;
+
+  if (res == 0)
+    {
+      set_info_from_stat (op_backend, info, &st, NULL, matcher);
+      
+      g_vfs_job_succeeded (G_VFS_JOB (job));
+    }
+  else
+    g_vfs_job_failed_from_errno (G_VFS_JOB (job), saved_errno);
+
+}
+
 static void
 do_close_read (GVfsBackend *backend,
 	       GVfsJobCloseRead *job,
@@ -812,15 +858,24 @@ do_create (GVfsBackend *backend,
   SMBCFILE *file;
   SmbWriteHandle *handle;
   smbc_open_fn smbc_open;
+  int errsv;
 
   uri = create_smb_uri (op_backend->server, op_backend->share, filename);
   smbc_open = smbc_getFunctionOpen (op_backend->smb_context);
+  errno = 0;
   file = smbc_open (op_backend->smb_context, uri,
-					O_CREAT|O_WRONLY|O_EXCL, 0666);
+		    O_CREAT|O_WRONLY|O_EXCL, 0666);
   g_free (uri);
 
   if (file == NULL)
-    g_vfs_job_failed_from_errno (G_VFS_JOB (job), errno);
+    {
+      errsv = fixup_open_errno (errno);
+
+      /* We guarantee EEXIST on create on existing dir */
+      if (errsv == EISDIR)
+	errsv = EEXIST;
+      g_vfs_job_failed_from_errno (G_VFS_JOB (job), errsv);
+    }
   else
     {
       handle = g_new0 (SmbWriteHandle, 1);
@@ -848,12 +903,13 @@ do_append_to (GVfsBackend *backend,
 
   uri = create_smb_uri (op_backend->server, op_backend->share, filename);
   smbc_open = smbc_getFunctionOpen (op_backend->smb_context);
+  errno = 0;
   file = smbc_open (op_backend->smb_context, uri,
 					O_CREAT|O_WRONLY|O_APPEND, 0666);
   g_free (uri);
 
   if (file == NULL)
-    g_vfs_job_failed_from_errno (G_VFS_JOB (job), errno);
+    g_vfs_job_failed_from_errno (G_VFS_JOB (job), fixup_open_errno (errno));
   else
     {
       handle = g_new0 (SmbWriteHandle, 1);
@@ -922,8 +978,9 @@ open_tmpfile (GVfsBackendSmb *backend,
     tmp_uri = g_strconcat (dir_uri, filename, NULL);
 
     smbc_open = smbc_getFunctionOpen (backend->smb_context);
+    errno = 0;
     file = smbc_open (backend->smb_context, tmp_uri,
-				       O_CREAT|O_WRONLY|O_EXCL, 0666);
+		      O_CREAT|O_WRONLY|O_EXCL, 0666);
   } while (file == NULL && errno == EEXIST);
 
   g_free (dir_uri);
@@ -969,13 +1026,13 @@ copy_file (GVfsBackendSmb *backend,
   smbc_close = smbc_getFunctionClose (backend->smb_context);
 
   from_file = smbc_open (backend->smb_context, from_uri,
-					  O_RDONLY, 0666);
+			 O_RDONLY, 0666);
   if (from_file == NULL || g_vfs_job_is_cancelled (job))
     goto out;
   
   to_file = smbc_open (backend->smb_context, to_uri,
-					O_CREAT|O_WRONLY|O_TRUNC, 0666);
-
+		       O_CREAT|O_WRONLY|O_TRUNC, 0666);
+  
   if (from_file == NULL || g_vfs_job_is_cancelled (job))
     goto out;
 
@@ -1045,11 +1102,12 @@ do_replace (GVfsBackend *backend,
   smbc_open = smbc_getFunctionOpen (op_backend->smb_context);
   smbc_stat = smbc_getFunctionStat (op_backend->smb_context);
   
+  errno = 0;
   file = smbc_open (op_backend->smb_context, uri,
-					O_CREAT|O_WRONLY|O_EXCL, 0);
+		    O_CREAT|O_WRONLY|O_EXCL, 0);
   if (file == NULL && errno != EEXIST)
     {
-      int errsv = errno;
+      int errsv = fixup_open_errno (errno);
 
       g_set_error_literal (&error, G_IO_ERROR,
 			   g_io_error_from_errno (errsv),
@@ -1114,11 +1172,12 @@ do_replace (GVfsBackend *backend,
 	      backup_uri = NULL;
 	    }
 	  
+	  errno = 0;
 	  file = smbc_open (op_backend->smb_context, uri,
-						O_CREAT|O_WRONLY|O_TRUNC, 0);
+			    O_CREAT|O_WRONLY|O_TRUNC, 0);
 	  if (file == NULL)
 	    {
-              int errsv = errno;
+              int errsv = fixup_open_errno (errno);
 
 	      g_set_error_literal (&error, G_IO_ERROR,
 				   g_io_error_from_errno (errsv),
@@ -1224,6 +1283,34 @@ do_seek_on_write (GVfsBackend *backend,
   return;
 }
 
+static void
+do_query_info_on_write (GVfsBackend *backend,
+			GVfsJobQueryInfoWrite *job,
+			GVfsBackendHandle _handle,
+			GFileInfo *info,
+			GFileAttributeMatcher *matcher)
+{
+  GVfsBackendSmb *op_backend = G_VFS_BACKEND_SMB (backend);
+  struct stat st = {0};
+  SmbWriteHandle *handle = _handle;
+  int res, saved_errno;
+  smbc_fstat_fn smbc_fstat;
+
+  smbc_fstat = smbc_getFunctionFstat (op_backend->smb_context);
+  res = smbc_fstat (op_backend->smb_context, handle->file, &st);
+  saved_errno = errno;
+
+  if (res == 0)
+    {
+      set_info_from_stat (op_backend, info, &st, NULL, matcher);
+      
+      g_vfs_job_succeeded (G_VFS_JOB (job));
+    }
+  else
+    g_vfs_job_failed_from_errno (G_VFS_JOB (job), saved_errno);
+
+}
+
 static void
 do_close_write (GVfsBackend *backend,
 		GVfsJobCloseWrite *job,
@@ -1388,15 +1475,17 @@ set_info_from_stat (GVfsBackendSmb *backend,
 					G_FILE_ATTRIBUTE_STANDARD_ICON))
     {
       icon = NULL;
+      content_type = NULL;
+      
       if (S_ISDIR(statbuf->st_mode))
 	{
 	  content_type = g_strdup ("inode/directory");
-	  if (strcmp (basename, "/") == 0)
+	  if (basename != NULL && strcmp (basename, "/") == 0)
 	    icon = g_themed_icon_new ("folder-remote");
 	  else
 	    icon = g_themed_icon_new ("folder");
 	}
-      else
+      else if (basename != NULL)
 	{
 	  content_type = g_content_type_guess (basename, NULL, 0, NULL);
 	  if (content_type)
@@ -1994,12 +2083,14 @@ g_vfs_backend_smb_class_init (GVfsBackendSmbClass *klass)
   backend_class->open_for_read = do_open_for_read;
   backend_class->read = do_read;
   backend_class->seek_on_read = do_seek_on_read;
+  backend_class->query_info_on_read = do_query_info_on_read;
   backend_class->close_read = do_close_read;
   backend_class->create = do_create;
   backend_class->append_to = do_append_to;
   backend_class->replace = do_replace;
   backend_class->write = do_write;
   backend_class->seek_on_write = do_seek_on_write;
+  backend_class->query_info_on_write = do_query_info_on_write;
   backend_class->close_write = do_close_write;
   backend_class->query_info = do_query_info;
   backend_class->query_fs_info = do_query_fs_info;
